// MG4interface.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2017 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Functions to interface to matrix elements generated with MadGraph
#ifndef VINCIA_MG4interface_H
#define VINCIA_MG4interface_H

// Include Pythia and headers
#include "Pythia8/Basics.h"
#include "Pythia8/PythiaComplex.h"
#include "Pythia8/Event.h"
#include "Pythia8/Info.h"
#include "Pythia8/ParticleData.h"
#include "Pythia8/PythiaStdlib.h"
#include "Pythia8/Settings.h"
#include "Pythia8/StandardModel.h"
//Used for ostream_iterator, which is used to convert a vector of integers
//into a vector of doubles
#include <iostream>
#include <iterator>

// Include Vincia Standard Library
#include "Vincia/VinciaCommon.h"

// Function to initialize MadGraph Interface. 
// Returns false if MadGraph dummy library linked, true otherwise.

namespace Vincia {

  class MG4interface {

  public:

    // Constructor.
    MG4interface() { 
      isInitPtr=false; 
      isInit=false;
      checkIS=false;
    };
    
    // Destructor
    virtual ~MG4interface() {};

    // Set pointers to required PYTHIA 8 objects
    void initPtr(Couplings* couplingsPtrIn, ParticleData* particleDataPtrIn, 
                 Settings* settingsPtrIn, VinciaCommon* vinComPtrIn) {
      couplingsPtr    = couplingsPtrIn;
      particleDataPtr = particleDataPtrIn;
      settingsPtr     = settingsPtrIn;
      vinComPtr       = vinComPtrIn;
      isInitPtr       = true;
    }
    
    // Initialization routine. Must be called for anything to work.
    bool init();

    // Wrapper to set a parameter in a MadGraph common block
    void setParm(string which, double what);
    void setParm(string which, complex what);

    // Wrapper to MadGraph matrix elements (without and with polarization)
    // (should return -1 if relevant ME not implemented)
    double ME2(int nIn, vector<int> id, vector<Vec4> p);
    // Madgraph uses integers for polarizations
    double ME2(int nIn, vector<int> id, vector<Vec4> p, vector<int> pol);
    // Wrapper to translate from double-precision polarizations
    double ME2(int nIn, vector<int> id, vector<Vec4> p, vector<double> pol){
      vector<int> polInt(pol.begin(), pol.end());    
      copy(pol.begin(), pol.end(), back_inserter(polInt));
      return ME2(nIn, id, p, polInt);    
    }
    
    // Set verbose level
    void setVerbose(int verboseIn) {verbose = verboseIn;}

    // Set color depth
    void setColorDepth(int colorDepthIn) {colorDepth = colorDepthIn;}
    // Get color depth
    int getColorDepth() {return colorDepth;}

    // Function to swap couplings (eg to use Zbb matrix element for Zcc)
    void swapZud();
    void swapAud();
    
    // Function to set the coupling for W
    void setWcouplings(int id1, int id2);
    
    // Order if we have q qbar' q' qbar'' q'' qbar with one distinct pair
    vector<int> orderQQBqqbqqb(pair<int,int> posQQB, bool firstLast);
    vector<int> orderQQBqqbqqbQCD(pair<int,int> posQQB, bool firstLast);
    
    // Functions to get order for W decays
    vector<int> orderWfourQuarks(vector<int> ids, bool& gSwap, bool& gg);
    vector<int> orderWfourQuarks(vector<int> ids, bool& gSwap) {
      bool gg = false; return orderWfourQuarks(ids, gSwap, gg); }
    vector<int> orderWfourQuarks(vector<int> ids) {
      bool gSwap = false, gg = false; return orderWfourQuarks(ids, gSwap, gg); }
    vector<int> orderWsixQuarks(vector<int> ids, map<int,int>& flavours,
      bool& flav2ordChange);
    
    // Function to set the Yukawa couplings and masses
    void setYukawa(int id1, double m1);
    void setYukawa(int id1, int id2, double m1, double m2);
    void setYukawa(int id1, int id2, int id3,
                   double m1, double m2, double m3);
    
    // Set either Higgs-g for Higgs-q coupling to zero
    void setHinterferenceOff(bool heftIn);
    
    // Set the positions of the incoming guys
    void setIIpos(int pos1In, int pos2In) {
      pos1Sav = min(pos1In,pos2In);
      pos2Sav = max(pos1In,pos2In);
    }
    
    // Print helicity combinations
    void printHels(int HEL[12]) {
      cout << " Using iHELs =";
      for (int iHel=0; iHel<12; iHel++) if (HEL[iHel] > 0)
        cout << " " << HEL[iHel] << ",";
      cout << endl;
    }
    
    // Temp flag to see if we are check initial state
    // side in g tau -> g g tau
    bool checkIS;
    
  private:
    
    // Is initialized?
    bool isInitPtr, isInit;

    // Pointers to VINCIA and PYTHIA 8 objects
    Couplings*    couplingsPtr;
    ParticleData* particleDataPtr;
    Settings*     settingsPtr;
    VinciaCommon* vinComPtr;

    // Color Mode (0: leading color, 1: vincia color, 2: full color)
    int colorDepth;
    
    // Positions of the incoming guys
    int pos1Sav;
    int pos2Sav;
    
    // CKM matrix elements for W coupling
    double Vud, Vus, Vub;
    double Vcd, Vcs, Vcb;
    double Vtd, Vts, Vtb;
    double gwf1, gwf2;
    
    // Yukawa couplings for the different flavours
    // and heft coupling
    double vev;
    vector<double> yukCoup;
    vector<double> yukCoupSav;
    double heftCoupSav;

    // Verbose level
    int verbose;

  };

}
#endif

