// TimeAntShower.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2017 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the TimeAntennaShower class 
  
#ifndef VINCIATIMEANTSHOWER_H
#define VINCIATIMEANTSHOWER_H

#include "Pythia8/TimeShower.h"
#include "Vincia/VinciaCommon.h"
#include "Vincia/VinciaWeights.h"

#include "Vincia/AntennaFunctions.h"
#include "Vincia/Branchelementals.h"
#include "Vincia/Reconnections.h"
#include "Vincia/Resolution.h"
#include "Vincia/TrialGeneratorsFSR.h"
#include "Vincia/MatrixElementCorrections.h"
#include "Vincia/MergingHooks.h"
#include "Vincia/Tools.h"

// Declare the F77 3->2 clustering map for use below
extern"C" {
  void vc3to2_(int *MMAP, double PA[4], double PR[4], double PB[4],
    double PAHAT[4], double PBHAT[4], int *IFAIL);
}

namespace Vincia {
  
// Forward declarations
class SpaceAntShower;
class MatrixElementCorrections;
class VinciaMergingHooks;
class VinciaMerging;
  
// Classes contained in this file 
class ResScaleHook;
class TimeAntShower;

//**************************************************************************

// UserHooks that are is to set the scale in resonance decays

class ResScaleHook : public UserHooks {
 public:
  ResScaleHook() { };
  virtual ~ResScaleHook() { };
  // Start resonance showers at m
  bool canSetResonanceScale() { return true; }
  virtual double scaleResonance(int iRes, const Event& event) {
    return event[iRes].m(); }
};

//**************************************************************************

// The VinciaFSR class
// Main shower class for final-state (FF) antenna showers
// Inherits from TimeShower in Pythia 8 so can be used as alternative to 
// TimeShower.
  
class TimeAntShower : public Pythia8::TimeShower {

  // Allow VinciaPlugin and SpaceAntShower to access private information
  friend class VinciaPlugin;
  friend class SpaceAntShower;

 public:

  // ----------------------------------------------------------------------

  // CONSTRUCTOR AND DESTRUCTOR 

  // Constructor.
  TimeAntShower() : isInit(false), headerIsPrinted(false) { };

  // Destructor.
  virtual ~TimeAntShower() {};

  // -----------------------------------------------------------------------
    
  // FUNCTIONS INHERITED FROM PYTHIA 8 FOR INTERFACING 
    
  // Initialize alphaStrong and related pTmin parameters.
  virtual void init( BeamParticle* beamAPtrIn = 0, 
    BeamParticle* beamBPtrIn = 0);
    
  // Possible limitation of first emission.
  virtual bool limitPTmax( Event& event, double Q2Fac = 0., double Q2Ren = 0.);

  // Potential enhancement factor of pTmax scale for hardest emission.
  // Used if limitPTmax = true.
  virtual double enhancePTmax() {
    return 1.0;
  }
    
  // Top-level routine to do a full time-like shower in resonance decay.
  virtual int shower( int iBeg, int iEnd, Event& event, double pTmax, 
    int nBranchMax = 0);
    
  // Pythia8::TimeShower has method
  // virtual int TimeShower::showerQED(int iBeg, int iEnd, Event& event,
  //                                  double pTmax)
  // This method is NOT OVERLOADED because we don't do QED showers
    
  // Used for global recoil scheme in Pythia8
  virtual void prepareGlobal(Event&) {};
    
  // Prepare system for evolution.
  virtual void prepare( int iSys, Event& event, bool limitPTmaxIn);

  // Update antenna list after each ISR emission.
  virtual void update( int iSys, Event& event, bool hasWeakRad = false);
    
  // Select next pT in downwards evolution.
  virtual double pTnext( Event& event, double pTbegAll, double pTendAll,
    bool isFirstTrial = false, bool doTrialIn = false);
    
  // Branch event
  virtual bool branch( Event& event, bool isInterleaved=false); 
    
  // Update antenna list after a multiple interactions rescattering.
  virtual void rescatterUpdate( int iSys, Event& event);
    
  // Only called if rescattering switched on in MPI (not supported by Vincia)
  virtual bool rescatterPropagateRecoil(Event&, Vec4&) {
    return false;
  };
    
  // Utility to print dipole list; for debug mainly.
  virtual void list( bool includeSaved ) const;
  // Compatibility with PYTHIA
  virtual void list() const { list(false); }
    
  // Functions inherited for merging.

  // Return clustered event.
  vector<Event> clusteredAll( const Event&, int, int, int, string);
  Event clustered( const Event& event, int iRa, int iEm, int iRe, string
    name ) {
    vector<Event> clusteredEvents = clusteredAll(event, iRa, iEm, iRe, name);
    return ( clusteredEvents.size() > 0 ? clusteredEvents[0] : Event() );
  }

  // Return the evolution variable.
  vector< map<string, double> > getStateVariablesAll( const Event&, int, int,
    int, string);
  map<string, double> getStateVariables( const Event& event, int iRa, int iEm, 
    int iRe, string name) { 
    vector< map<string, double> > scales = getStateVariablesAll(event, iRa, iEm,
      iRe, name);
    return ( scales.size() > 0 ? scales[0] : map<string, double>() );
  }

  // Check if attempted clustering is handled by timelike shower.
  virtual bool isTimelike( const Event&, int, int, int, string);

  // Return a string identifier of a splitting.
  vector<string> getSplittingName( const Event&, int, int, int);

  // Return the splitting probability.
  vector<double> getSplittingProbAll( const Event&, int, int, int, string);
  double getSplittingProb( const Event& event, int iRa, int iEm, int iRe, 
    string name) { 
    vector<double> splittingProbs = getSplittingProbAll(event, iRa, iEm, iRe, 
      name);
    return ( splittingProbs.size() > 0 ? splittingProbs[0] : 0.0 );
  }

  // TODO
  //bool allowedSplitting( const Event&, int, int);

  // TODO
  //vector<int> getRecoilers( const Event&, int, int, string);

  // END OF FUNCTIONS INHERITED FROM PYTHIA 8 FOR INTERFACING

  // -----------------------------------------------------------------------

  // PUBLIC NON-INHERITED MEMBERS

  // Return matrix-element correction.
  double getMEC( const Event& event, int nStep=-1);

  // Return scales, alphaS values, etc in case of only unordered shower paths.
  map<string, double> getUnorderedInfos( const Event& event);

  // Return the Born probability (= ME).
  double getBornProb( const Event& event);

  // Return the Born scale for merging.
  double getBornScale( const Event& event);

  // Return the shower starting scale for pure shower.
  double getShowerStartingScale( const Event& event);

  // Update antennae if ISR changed system in prepare
  void updateAfterISR(int iSys, Event& event, map<int,int> iChanged);
    
  // Wrapper function to return a specific antenna inside antenna set 
  AntennaFSR* getAnt(int iAnt) { return antSetPtr->getAnt(iAnt); }

  // Initialize Pointers to Antenna sets
  void initAntPtr(AntennaSetFSR* antSetIn) { antSetPtr = antSetIn; }

  // Print header information (version, settings, parameters, etc)
  void header();

  //Initialize internal and diagnostic histograms
  void initVinciaHistos();

  // Print final statistics information
  void printInfo( bool pluginCall = false );

  // Print internal and diagnostic histrograms
  void printHistos() {
    // Print histograms in shower class
    map<string,Hist>::iterator iH;
    for (iH=vinciaHistos.begin();iH!=vinciaHistos.end();++iH) {
      string Hname=iH->first;
      if ( vinciaHistos[Hname].getEntries() >= 1) {
        cout<<Hname<<vinciaHistos[Hname]<<endl;
      }
    }
  }
    
  // Print internal and diagnostic histrograms
  void writeHistos(string fileName="vincia", string suffix="dat") {
    // Print histograms in shower class
    map<string,Hist>::const_iterator iH;
    for (iH=vinciaHistos.begin();iH!=vinciaHistos.end();++iH) {
      string Hname=iH->first;
      if ( vinciaHistos[Hname].getEntries() >= 1) {
        string file = fileName + "-Hist-" + Hname + "." + suffix;
        file = sanitizeFileName(file);
        cout << "Writing " << file << endl;
        iH->second.table(file);
      }
    }
  }
    
  // Allow the user to get a pointer to an internal histogram
  // Note: points to an entry of a map, so must be updated each
  // time it is used.
  Hist* getHistPtr(string histIn) {
    if (vinciaHistos.find(histIn) != vinciaHistos.end()) 
      return &vinciaHistos[histIn];
    else return 0;
  }
    
  const Hist& getDiagnosticHistogram( string name ) {
    return vinciaHistos[name];
  }
    
  // Kinematic maximum for gluon emission scale
  double yMaxEmit() { return 1./evNormFactor; }

  // Evolution Windows: Phase space region boundaries
  int getRegion(double q) {
    if (q < mc/5) return 0;
    else if (q < mc) return 1;
    else if (q < mb) return 2;
    else if (q < mtb) return 3;
    else if (q < mt) return 4;
    else return int( log(q/mt)/log(5) + 5);
  }    

  // Evolution Windows: Phase space region boundaries
  double getQmin(int iRegion) {
    if (iRegion <= 0) return mc/16;
    else if (iRegion == 1) return mc/5;
    else if (iRegion == 2) return mc;
    else if (iRegion == 3) return mb;
    else if (iRegion == 4) return mtb;
    else if (iRegion == 5) return mt;
    else return pow(5,(double)iRegion-5.0)*mt;
  }

  // Number of active flavors 
  int getNf(int iRegion) {
    if (iRegion <= 1) return 3;
    else if (iRegion <= 2) return 4;
    else if (iRegion <= 4) return 5;
    else return 6;
  }

  // Pointer to AlphaS instance
  AlphaStrong* alphaSptr;

  // Lambda value
  double getLambda(int nFin) {
    if (nFin <= 3) return alphaSptr->Lambda3();
    else if (nFin <= 4) return alphaSptr->Lambda4();
    else if (nFin <= 5) return alphaSptr->Lambda5();
    else return alphaSptr->Lambda6();
  }

  // Functions to initialize, set and get headRoomFactors
  void initHeadRoom(double valEmit,  double valEmitMEC, 
    double valSplit, double valSplitMEC) {
    for (int iReg=0; iReg<=20; iReg++) {
      for (int nGin=0; nGin<=8; nGin++) {
        for (int nQin=0; nQin<=8; nQin++) {
          headRoom[0][iReg][nGin][nQin]     = valEmit;
          headRoom[1][iReg][nGin][nQin]     = valSplit;
          headRoomMECs[0][iReg][nGin][nQin] = ( valEmitMEC + min(1.0, 1.0/3.0*max(0,nGin+nQin-1)) );
          headRoomMECs[1][iReg][nGin][nQin] = valSplitMEC;
          headRoomOptimize[0][iReg][nGin][nQin]     = optimizeHeadRoom;
          headRoomOptimize[1][iReg][nGin][nQin]     = optimizeHeadRoom;
          headRoomMECsOptimize[0][iReg][nGin][nQin] = optimizeHeadRoom;
          headRoomMECsOptimize[1][iReg][nGin][nQin] = optimizeHeadRoom;
        }
      }
    }
  }
  double getHeadRoom(int iBranchType, int iReg, int nGin, int nQin,
    int iSys, double qMin) {
    if (!iRegDependence) iReg = 1;
    else if (iReg == -1) iReg = getRegion(qMin);
    if (iBranchType < 0 || iReg < 0  || nGin < 0 || nGin > 8 || 
      iBranchType > 1 || iReg > 20 || nQin < 0 || nQin > 8) return 1.0;
    if ((nBranch[iSys]+1<=matchingLO) && hardSys[iSys])
      return headRoomMECs[iBranchType][iReg][nGin][nQin];
    else return headRoom[iBranchType][iReg][nGin][nQin];
  }
  void setHeadRoom(int iBranchType, int iReg, int nGin, int nQin,
    int iSys, double qMin, double scale) {
    if (!iRegDependence) iReg = 1;
    else if (iReg == -1) iReg = getRegion(qMin);
    if (iBranchType < 0 || iReg < 0  || nGin < 0 || nGin > 8 || 
      iBranchType > 1 || iReg > 20 || nQin < 0 || nQin > 8) return;
    bool MEC      = ((nBranch[iSys]+1<=matchingLO) && hardSys[iSys]);
    double maxVal = (iBranchType == 0 ? 2.0 : 4.0);
    double oldVal = (MEC ? headRoomMECs[iBranchType][iReg][nGin][nQin]
      : headRoom[iBranchType][iReg][nGin][nQin]);
    // Switch off further optimization if increasing headroom factor
    if ( scale > 1.0 ) {
      if (MEC) {
        headRoomMECsOptimize[iBranchType][iReg][nGin][nQin] = false;
        headRoomMECs[iBranchType][iReg][nGin][nQin] = min(scale*oldVal,maxVal);
      } else {
        headRoomOptimize[iBranchType][iReg][nGin][nQin] = false;
        headRoom[iBranchType][iReg][nGin][nQin] = min(scale*oldVal,maxVal);
      }
      if (verbose >= 5) {
        printOut("TimeAntShower::setHeadRoom","Increasing headroom at nG = "
          +num2str(nGin,2)+" nQ = "+num2str(nQin,2)+"; current "
          "headRoom = "+num2str(oldVal));
      }
    } else {
      // Allow reduction of headroom if we're still optimizing
      if (MEC) {
        if (headRoomMECsOptimize[iBranchType][iReg][nGin][nQin])
          headRoomMECs[iBranchType][iReg][nGin][nQin] = scale*oldVal;
      } else {
        if (headRoomOptimize[iBranchType][iReg][nGin][nQin])
          headRoom[iBranchType][iReg][nGin][nQin] = scale*oldVal;
      }
    }
  }

  // Function to ask if a system is hard
  bool isSysHard(int iSys) {
    if (!isInit) return false;
    if ((int)hardSys.size() <= iSys) return false;
    return hardSys[iSys];
  }
    
  // Information.
  // ---------------------------------------------------------------------------
  // Get number of systems
  int getNsys() {return nBranchFSR.size();}
  // Get number of branchings in a system (return -1 if no such system)
  int getNbranch(int iSys = -1) {
    // iSys < 0 : return sum over all
    int n = 0;
    if (iSys < 0) for (int i=0; i<(int)nBranchFSR.size(); ++i)
                    n += nBranchFSR[i];
    else if (iSys < (int)nBranchFSR.size()) 
      n = nBranchFSR[iSys];
    else n = -1;
    return n;
  }
  // Get scale of branchings; use (0,1) for first branching in 1st system.
  // (Could be extended so that (i,0) would return starting scale for system i.)
  double getQbranch(int iSys, int iBranch) {
    if (iSys < 4 && iSys >= 0 && iBranch <= 10 && iBranch >= 1)
      return qBranch[iSys][iBranch];
    else return 0.;
  }
  double getPTphys(int iSys, int iBranch) {
    if (iSys < 4 && iSys >= 0 && iBranch <= 10 && iBranch >= 1)
      return pTphys[iSys][iBranch];
    else return 0.;
  }

  // Set verbose level
  void setVerbose(int verboseIn) { verbose = verboseIn; }

 private:
    
  // -----------------------------------------------------------------------
    
  // PRIVATE MEMBERS
    
  // Generate trial branching kinematics and check physical phase space
  bool generateKinematics(BranchelementalFSR* trialPtr);

  // Main trial accept function
  bool acceptTrial(Event& event, BranchelementalFSR* trialPtr, 
    int iSysBranch, vector<Particle> partsNow, vector<int> antPos);

  // Method to assign colour flow
  bool assignColourFlow(Event& event, BranchelementalFSR* trialPtr);

  // Veto function for sector ordering.
  bool sectorAccept(Event& event, BranchelementalFSR* trialPtr);

  // Initialized?
  bool isInit;
    
  // Beams
  BeamParticle* beamAPtr;
  BeamParticle* beamBPtr;
  double sbbSav;
    
  // Possibility to allow user veto of emission step.
  bool hasUserHooks, canVetoEmission;

  // Parameter setting which kind of 2->4 modifications (if any) are used
  int mode2to4;
  
  // Shower parameters
  int evolutionType, nGluonToQuarkF, cutoffType;
  double cutoffScale, relativisticSpeed;
  double evNormFactor, normFacQ2max;
  int nFlavZeroMassDyn, nFlavZeroMassKin;
  int massCorrectionEmit, massCorrectionSplit;
  bool treatAsMassless;
    
  // AlphaS parameters 
  bool useCMW;
  int alphaSorder;
  double alphaSvalue, alphaSmax, alphaSmuFreeze, alphaSmuMin;
  double aSkMu2, aSkMu2Split;
  double mt, mtb, ms, mb, mc;
  // Calculated values
  double mu2freeze, mu2min;

  // Trial generators
  map<string,TrialGeneratorFSR*> trialGenPtrs;
  TrialSoft trialSoft;
  TrialCollI trialCollI;
  TrialCollK trialCollK;
  TrialSplitI trialSplitI;
  TrialSplitK trialSplitK;
  //TrialConstEmit trialConstEmit;
  //TrialConstSplit trialConstSplit;

  // Uncertainty Bands and Weights
  bool uncertaintyBands, allowReweighting;

  // Colour Reconnection parameters
  int  CRmode;
  bool CRjunctions;

  // Sector 
  int  sectorType;
  bool sectorShower, useSectorTerms;

  // Polarization global flag
  bool helicityShower;
  // Polarization and mass local flags
  bool isMassless, usePolarizationNow;
    
  // Desired matching multiplicities, level, regulator, regulator scales
  MatrixElementCorrections* matrixElementCorrectionsPtr;
  bool meCorrections, meCorrectionsNow;
  int  matchingLO, matchingNLO;
  int  matchingRegOrder, matchingRegShape;
  double matchingScale, matchingScaleRatio, matchingIRcutoff;
  bool matchingScaleIsAbs, matchingFullColor;
  bool Hinterference;
  // Interfaces to matrix elements
  bool useMG4, useMG5, useMHV;
  // Hyperjet on/off (high-y-per-jet: experimental interface to HEJ)
  bool hyperjet;
  
  // Pointer to Evolution Variable
  Resolution* resolutionPtr;

  // Pointer to VINCIA common, functions, clustering
  VinciaCommon* vinComPtr;
  Functions* functionsPtr;
  VinClu* vinCluPtr;

  // Pointer to weights class
  VinciaWeights* weightsPtr;

  // Pointer to VINCIA ISR shower
  SpaceAntShower* ISRshowerPtr;

  // Pointer to VINCIA Reconnections object
  Reconnections* reconnectionsPtr;

  // New headRoomFactors
  // first arg = emi/split, second arg = iRegion, third arg = nG, fourth arg = nG
  double headRoom[2][21][9][9];
  bool   headRoomOptimize[2][21][9][9];
  double headRoomMECs[2][21][9][9];
  bool   headRoomMECsOptimize[2][21][9][9];
  bool   optimizeHeadRoom, iRegDependence;

  // Vector of antenna sets
  AntennaSetFSR* antSetPtr;   

  // Total and MEC accept probability
  vector<double> Paccept;
  double PME;

  // Factorization scale and shower starting settings
  int    qMaxMatch;
  double q2fac, qFac;
  double q2maxFudge, q2maxFudgeMPI;

  // Parameters for merging and pointer to Vincia's merging
  bool doMergeSav;
  int  incomingType;
  VinciaMergingHooks* vinMergingHooksPtr;

  // current winner
  BranchelementalFSR* winnerPtr;
  int iSysWin;
    
  // ---------------------------- Every system gets own copy -------------------------------- //
    
  // Is system a hard system?
  vector<bool> hardSys;
    
  // Some properties for MECs
  vector<int> MEclass, nQLneeded, Hint, additionalJets;

  // Saved particle state and number in event record
  vector< vector< Particle > > partsSav;
  vector< vector< int      > > indexSav;

  // Polarization local flags
  vector<bool> polarizedSys;

  // vector of dipoles (with trial branchings) 
  vector< vector<BranchelementalFSR> > branchelementals;
  vector< BranchelementalFSR > savedBranchings;
    
  // Save initial FSR starting scale system by system
  vector<double> Q2hat;
    
  // Count the number of branchings in the system
  vector<int> nBranch, nBranchFSR;
    
  // Total mass of showering system
  vector<double> mSystem;
    
  // Is the system a resonance decay?
  vector<bool> isResonance;
    
  // Count numbers of quarks and gluons
  vector<int> nG, nQQ;
    
  // ---------------------------- //

  // Internal histograms
  map<string,Hist> vinciaHistos;
    
  // Bool to know whether we have binned the first QBranch already
  bool firstQBranchBinned;
    
  // Statistics
  int nTrialsSum, nTrialsVeto;
  int nTrialsHull, nTrialsHullMassive;
  int nTrialsCutM, nTrialsCutQ;
  int nClosePSforHQ;
  vector<int> nMCviols, nNegPas;
  vector<double> maxMCviols, minNegPas;

  // Verbose setting, P>1 warning settings, and version number
  int verbose;
  double pAcceptFuss, pAcceptPT0;
  int nReportMCviol, nReportedMCviol;
  double version;
  bool headerIsPrinted;
  double qBranch[4][11], pTphys[4][11];
    
  // Counter for numbers of events
  long nAccepted, nSelected;
  int nVetoUserHooks, nFailHadLevel, nCallPythiaNext;

};
 
}

#endif
