// Resolution.cc is a part of the VINCIA plug-in to the 
// PYTHIA 8 generator. 
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2017 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the Resolution 
// class. Note: also used for sector decomposition variables.

// Include the Resolution header
#include "Vincia/Resolution.h"

namespace Vincia {
    
  // Initialize
  bool Resolution::init() {
    // Check that pointers initialized
    if (! isInitPtr) {
      printOut("Resolution::init","Cannot initialize, pointers not set.");
      return false;
    }
    // Set verbose level
    verbose         = settingsPtr->mode("Vincia:verbose");
    // Normalization factors for FSR
    pTnormalization = settingsPtr->parm("Vincia:pTnormalization"); 
    mDnormalization = settingsPtr->parm("Vincia:mDnormalization"); 
    // Normalization factors for ISR
    pT2IInorm       = settingsPtr->parm("Vincia:pTnormalizationII");
    pT2IFnorm       = settingsPtr->parm("Vincia:pTnormalizationIF");
    MsplitInorm     = settingsPtr->parm("Vincia:mSnormalizationI");
    MconvInorm      = settingsPtr->parm("Vincia:mCnormalizationI");
    MsplitFnorm     = settingsPtr->parm("Vincia:mSnormalizationF");
    // Set these initially to zero
    isInit          = true;
    return isInit;
  }

  // Definition of the dimensionless resolution function
  double Resolution::yE(double y12, double y23, int typeIn) {
    
    // Default type == evolution variable
    if (typeIn == 0) typeIn = settingsPtr->mode("Vincia:evolutionType");
    double yENow = 1.;

    // ----------------------------------------------------------------
    // Evolution / Phase-space ordering variables, by type : 
    
    if (typeIn == 1) {
      // Type 1 : pT-ordering (proportional to ARIADNE pT)
      yENow= pTnormalization*y12*y23;    
    } 

    else if (typeIn == 2) {
      // Type 2 : daughter mass-ordering (mAnt-ordering)
      yENow = mDnormalization * min(y12,y23);
    } 
    
    else if (typeIn == 3) {
      // Type 3 : energy ordering 
      // Warning: not infrared safe a priori, but combined with
      // hadronization cut in pT can still be used.
      yENow= pow2(y12 + y23);
    } 

    else if (typeIn == 4) {
      // Type 4 : anticollinear (Upsilon) ordering
      yENow= pow(pow2(y12+y23),0.25)-pow(pow2(y12-y23),0.25);
    }

    else if (typeIn == 5) {
      // Type 5 : collinear-safe energy ordering 
      yENow= sqrt(pow2(pow2(y12+y23))-pow2(pow2(y12-y23)));
    }

    else if (typeIn == -1 || typeIn == 6) {
      // Type 6 and -1 : Pythia pTevol = min(pTevolA,pTevolB)
      double yEvolA = y12*(1-y23)*(y12+y23)/pow2(1-y12); 
      double yEvolB = y23*(1-y12)*(y23+y12)/pow2(1-y23); 
      yENow= min(yEvolA,yEvolB);
    }
  
    else if (typeIn == 7) {
      // Type 7: m2jk ordering (for X g -> X qbar q)
      yENow = y23;
    }
    
    //////////////
    
    // Sector shower 
    // NOTE: THE FOLLOWING ARE NOT EVOLUTION VARIABLES IN THE PROPER SENSE, 
    // BUT SECTOR DECOMPOSITION VARIABLES. 

    else if (typeIn == 8) {
      // Type 8: sector shower for quarks with color-ordered 
      // clustering Xg -> X1 qbar2 q3 (i.e., parton 2 = antiquark) 
      yENow= sqrt(y12)*y23;
    }

    else if (typeIn == 9) {
      // Type 9: sector shower for quarks with color-ordered 
      // clustering gX -> qbar1 q2 X (i.e., parton 2 = quark)
      yENow= y12*sqrt(y23);
    }

    else if (typeIn == 10) {
      // Type 10: sector shower for gluons 
      // (ARIADNE PT)
      yENow= y12*y23;
    }

    //////////////
    
    else if (typeIn == 11) {
      // Type 11 : phase-space ordering / virtuality-ordering 
      // Warning: not infrared safe
      double yMax = max(y12,y23);
      yENow= yMax;
    } 

    else if (typeIn == 12) {
      // Type 12 : experimental Rapidity-like ordering 
      // (warning: not infrared safe)
      // Divide phase space into y12 > y23 and y12 < y23
      double yRap=0.0;
      if (y12 < y23) {
        double ct23 = 1-y23/2.0/(1-y12)/(y12+y23);
        double trap = M_PI - acos(ct23);
        yRap = sqrt((1.0-cos(trap))/(1.0+cos(trap)));      
      } else {
        double ct12 = 1-y12/2.0/(1-y23)/(y12+y23);
        double trap = M_PI - acos(ct12);
        yRap = sqrt((1.0-cos(trap))/(1.0+cos(trap)));      
      }
      yENow= yRap;
    } 

    else if (typeIn == 13) {
      // Type 13 : Thrust-ordering. 
      // Warning: not infrared safe    
      yENow= min(min(y12,y23),1.-y12-y23);
    } 

    else if (typeIn == 15) {
      // Type 15 : experimental angular-like ordering (warning: not infrared safe)
      // Divide phase space into y12 > y23 and y12 < y23
      if (y12 < y23) {
        yENow= max(y12/(1-y23),y12/y23);      
      } else {
        yENow= max(y23/(1-y12),y23/y12);      
      }
    } 

    else {
      // Print warning if type not recognized at all
      if (verbose >= 1)
        printErr("Vincia::Resolution::R",
                 "Unknown ordering variable :"+num2str(typeIn));
      yENow= -1.;
    }

    // Return
    return yENow;

  }
  
  // Copy of R using 3 Vec4 as input 
  double Resolution::yE(Vec4 v1, Vec4 v2, Vec4 v3, int typeIn) {
    // Compute invariants (including masses)
    // MR 12.8.2010: changed to sij=2*pi*pj and s = (p1+p2+p3)^2
    double s=m2(v1,v2,v3);
    if (s == 0.0) return 0.0;
    double y12=(2*v1*v2)/s;
    double y23=(2*v2*v3)/s;
    return yE(y12,y23,typeIn);    
  }

  // Copy of R using 3 double[4] as input 
  double Resolution::yE(double* p1,double* p2,double* p3, int typeIn) {
    // Translate from doubles to Vec4, then call Vec4 function.
    Vec4 v1(p1[0],p1[1],p1[2],p1[3]),v2(p2[0],p2[1],p2[2],p2[3]),
      v3(p3[0],p3[1],p3[2],p3[3]);
    return yE(v1,v2,v3,typeIn);
  }

  // Copy for returning R*s
  double Resolution::Q2E(Vec4 v1, Vec4 v2, Vec4 v3, int typeIn) { 
    // Compute invariants (including masses)
    // MR 12.8.2010: changed to sij=2*pi*pj and s = (p1+p2+p3)^2
    double s=m2(v1,v2,v3);
    if (s == 0.0) return 0.0;
    double y12=(2*v1*v2)/s;
    double y23=(2*v2*v3)/s;
    return yE(y12,y23,typeIn)*s;
    
  }
  
  // Copy for returning R*s from Particles
  double Resolution::Q2E(Particle p1, Particle p2, Particle p3, int typeIn) {
    return Q2E(p1.p(),p2.p(),p3.p(),typeIn);
  }

  // Copy for returning R*s from y12, y23, and s
  double Resolution::Q2E(double s, double y12, double y23, int typeIn) {
    return s*yE(y12,y23,typeIn);
  }
  
  // Ordering variable (dimensionful) for ISR -- II
  double Resolution::Q2EII(double sAB, double saj, double sjb, int typeIn) {
    // Evolution / Phase-space ordering variables, by type:
    if (typeIn == 1) { // Gluon emission
      return pT2IInorm*saj*sjb/(sAB+saj+sjb); // pT2
    } 
    else if (typeIn == 2) { // Gluon Splitting
      return MsplitInorm*saj; // Mass2
    }
    else if (typeIn == 3) { // Gluon Conversion
      return MconvInorm*saj; // Mass2
    } 
    else return 0.0;
  }
  
  double Resolution::Q2EII(Vec4 pa, Vec4 pb, Vec4 pj, int typeIn) {
    double saj = 2.0*dot4(pa,pj);
    double sjb = 2.0*dot4(pb,pj);
    double sab = 2.0*dot4(pa,pb);
    double sAB = sab - saj - sjb;
    return Q2EII(sAB, saj, sjb, typeIn);
  }
  
  double Resolution::Q2EII(Particle pa, Particle pb, Particle pj, int typeIn) {
    double saj = 2.0*dot4(pa,pj);
    double sjb = 2.0*dot4(pb,pj);
    double sab = 2.0*dot4(pa,pb);
    double sAB = sab - saj - sjb;
    return Q2EII(sAB, saj, sjb, typeIn);
  }
  
  // Ordering variable (dimensionful) for ISR -- IF
  double Resolution::Q2EIF(double sAK, double saj, double sjk, int typeIn) {
    // Evolution / Phase-space ordering variables, by type:
    if (typeIn == 1) { // pT-ordering for gluon emission
      return pT2IFnorm*saj*sjk/(sAK+sjk); 
    } 
    else if (typeIn == 2) { // Gluon Splitting in I
      return MsplitInorm*saj; // Mass2
    }
    else if (typeIn == 3) { // Gluon Conversion
      return MconvInorm*saj; // Mass2
    }
    else if (typeIn == 4) { // Gluon Splitting in F
      return MsplitFnorm*sjk; // Mass2
    }
    else return 0.0;
  }
  
  double Resolution::Q2EIF(Vec4 pa, Vec4 pk, Vec4 pj, int typeIn) {
    double saj = 2.0*dot4(pa,pj);
    double sjk = 2.0*dot4(pk,pj);
    double sak = 2.0*dot4(pa,pk);
    double sAK = sak + saj - sjk;
    return Q2EIF(sAK, saj, sjk, typeIn);
  }
  
  double Resolution::Q2EIF(Particle pa, Particle pk, Particle pj, int typeIn) {
    double saj = 2.0*dot4(pa,pj);
    double sjk = 2.0*dot4(pk,pj);
    double sak = 2.0*dot4(pa,pk);
    double sAK = sak + saj - sjk;
    return Q2EIF(sAK, saj, sjk, typeIn);
  }

}
