// VinciaWeights.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator. 
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2017 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

#include "Vincia/VinciaWeights.h"

namespace Vincia {

  //==========================================================================

  // The VinciaWeights class.

  const double VinciaWeights::TINYANT       = 1.0e-10;
  const double VinciaWeights::PACCEPTVARMAX = 0.99;
  const double VinciaWeights::MINVARWEIGHT  = 0.01;

  //--------------------------------------------------------------------------

  // Initialize.

  void VinciaWeights::init() {

    // Check initPtr
    if (!isInitPtr) {
      printOut("VinciaWeights::init","Error! pointers not initialized"); 
      return;
    }

    verbose          = settingsPtr->mode("Vincia:verbose");
    allowReweighting = settingsPtr->flag("Vincia:allowReweighting");
    uncertaintyBands = settingsPtr->flag("Vincia:uncertaintyBands");
    varLabels        = settingsPtr->wvec("Vincia:uncertaintyBandsList");
    nWeightsSav      = (uncertaintyBands ? 1+varLabels.size() : 1);

    // List of all keywords
    allKeywords.clear(); allKeywords.resize(0);
    string ffKeys[5] = { ":", ":qqemit:", ":qgemit:", ":ggemit:", ":xgsplit:" };
    string ifKeys[8] = { ":", ":qqemit:", ":gqemit:", ":ggemit:", ":ggemit:",
      ":qxsplit:", ":gxconv:", ":xgsplit:" };
    string iiKeys[6] = { ":", ":qqemit:", ":gqemit:", ":ggemit:", ":qxsplit:",
      ":gxconv:" };
    for (int i=0; i<5; i++) allKeywords.push_back("ff"+ffKeys[i]+"murfac");
    for (int i=0; i<5; i++) allKeywords.push_back("ff"+ffKeys[i]+"cns");
    for (int i=0; i<8; i++) allKeywords.push_back("if"+ifKeys[i]+"murfac");
    for (int i=0; i<8; i++) allKeywords.push_back("if"+ifKeys[i]+"cns");
    for (int i=0; i<6; i++) allKeywords.push_back("ii"+iiKeys[i]+"murfac");
    for (int i=0; i<6; i++) allKeywords.push_back("ii"+iiKeys[i]+"cns");

    // Convert iAntPhys to keyword
    iAntToKeyFSR.clear();
    iAntToKeyFSR[iQQemit]  = "qqemit";
    iAntToKeyFSR[iQGemit]  = "qgemit";
    iAntToKeyFSR[iGGemit]  = "ggemit";
    iAntToKeyFSR[iGGsplit] = "xgsplit";
    iAntToKeyFSR[iQGsplit] = "xgsplit";
    iAntToKeyISR.clear();
    iAntToKeyISR[iQQemitII]  = "qqemit";
    iAntToKeyISR[iGQemitII]  = "gqemit";
    iAntToKeyISR[iGGemitII]  = "ggemit";
    iAntToKeyISR[iQXsplitII] = "qxsplit";
    iAntToKeyISR[iGXconvII]  = "gxconv";
    iAntToKeyISR[iQQemitIF]  = "qqemit";
    iAntToKeyISR[iQGemitIF]  = "qgemit";
    iAntToKeyISR[iGQemitIF]  = "gqemit";
    iAntToKeyISR[iGGemitIF]  = "ggemit";
    iAntToKeyISR[iQXsplitIF] = "qxsplit";
    iAntToKeyISR[iGXconvIF]  = "gxconv";
    iAntToKeyISR[iXGsplitIF] = "xgsplit";

    // Clean up the names of requested variations.
    for (int i=0; i<(int)varLabels.size(); i++) {
      int strBegin = varLabels[i].find_first_not_of(" ");
      if ((i == 0) && (varLabels[i].find("{") != string::npos)) strBegin += 1;
      int strEnd   = varLabels[i].find_last_not_of(" ");
      if ((i == (int)varLabels.size()-1) && (varLabels[i].find("}") !=
        string::npos)) strEnd -= 1;
      int strRange = strEnd - strBegin + 1;
      varLabels[i] = toLower(varLabels[i].substr(strBegin, strRange));
    }

    // Parse names of requested variations and check for keywords
    varKeys.clear(); varKeys.resize(varLabels.size());
    varVals.clear(); varVals.resize(varLabels.size());
    for (int i=0; i<(int)varLabels.size(); i++) {
      varKeys[i].clear(); varKeys[i].resize(0);
      varVals[i].clear(); varVals[i].resize(0);
      string var      = varLabels[i];
      int    iEndName = var.find(" ", 0);
      string varName  = var.substr(0, iEndName);
      string left     = var.substr(iEndName, var.length());
             left     = left.substr(left.find_first_not_of(" "), left.length());
      varLabels[i] = varName;
      while (left.length() > 1) {
        int    iEnd = left.find("=", 0);
        int    iAlt = left.find(" ", 0);
        if ( (iAlt < iEnd) && (iAlt > 0) ) iEnd = iAlt;
        string key  = left.substr(0, iEnd);
        if (1+key.find_last_not_of(" ") < key.length())
          key = key.substr(0, 1+key.find_last_not_of(" "));
        varKeys[i].push_back(key);
        string val  = left.substr(iEnd+1, left.length());
               val  = val.substr(val.find_first_not_of(" "), val.length());
               val  = val.substr(0, val.find_first_of(" "));
        varVals[i].push_back(atof(val.c_str()));
               left = left.substr(iEnd+1, left.length());
        if (left.find_first_of(" ") >= left.length()) break;
               left = left.substr(left.find_first_of(" "), left.length());
        if (left.find_first_not_of(" ") >= left.length()) break;
               left = left.substr(left.find_first_not_of(" "), left.length());
      }
    }

    if (uncertaintyBands && (verbose >= 4)) {
      printOut("VinciaWeights", "List of variations, keywords and values:");
      for (int i=0; i<(int)varLabels.size(); i++) {
        cout << "  " << varLabels[i] << " : ";
        for (int j=0; j<(int)varVals[i].size(); j++) {
          cout << " " << varKeys[i][j] << " -> " << varVals[i][j];
          if (j<(int)varVals[i].size()-1) cout << ",";
        }
        cout << endl;
      }
    }

    // Safety check for non-sensible input
    if (uncertaintyBands) {
      for (int i=0; i<(int)varKeys.size(); i++) {
        for (int j=0; j<(int)varKeys[i].size(); j++) {
          // Check input keywords against known ones
          bool foundValidKey = false;
          for (int k=0; k<(int)allKeywords.size(); k++) {
            if (allKeywords[k] == varKeys[i][j]) {
              foundValidKey = true;
              break;
            }
          }
          if (!foundValidKey)
            printOut("VinciaWeights", "Error! Unknown key " + 
              varKeys[i][j] + " found, please check!");
        }
      }
    }

    deltaASmax        = settingsPtr->parm("Vincia:deltaAlphaSmax");
    cNSpTmin          = settingsPtr->parm("Vincia:cNSpTmin");
    varInMPIshowers   = settingsPtr->flag("Vincia:uncertaintyInMPIshowers");
    muSoftCorr        = settingsPtr->flag("Vincia:muSoftCorr");
    printPaccVarViol  = settingsPtr->flag("Vincia:PrintPacceptVariations");
    printReWeightViol = settingsPtr->flag("Vincia:PrintReWeightVariations");
    allowNegAnt       = false;

    nReportWeight      = 5;
    nReportedWeight    = 0;

    // Resize weight arrays to the relevant number of elements
    weights.resize(nWeightsSav);
    weightsOld.resize(nWeightsSav);
    weightsMax.resize(nWeightsSav);
    weightsMin.resize(nWeightsSav);
    weightSum.resize(nWeightsSav);
    weightSum2.resize(nWeightsSav);
    contribSum.resize(nWeightsSav);
    contribSum2.resize(nWeightsSav);
      
    // Initialize counters and flags
    nTotWeights               = 0;
    nNonunityWeight           = 0;
    nNegativeWeight           = 0;
    nNonunityInitialWeight    = 0;
    nNegativeInitialWeight    = 0;
    nNonunityWeightNow        = 0;
    nNegativeWeightNow        = 0;
    nNonunityInitialWeightNow = 0;
    nNegativeInitialWeightNow = 0;
    didReweight               = false;
    firstCall                 = false;

    // Initialize weights to 0 (or 1 respectively)
    for (int iVar=0; iVar<nWeightsSav; iVar++) {
      weights[iVar]     = 1.0;
      weightsOld[iVar]  = 0.0;
      weightsMax[iVar]  = 0.0;
      weightsMin[iVar]  = 1.0e10;
      weightSum[iVar]   = 0.0;
      weightSum2[iVar]  = 0.0;
      contribSum[iVar]  = 0.0;
      contribSum2[iVar] = 0.0;
    }

  }

  //--------------------------------------------------------------------------

  // Reset the weights, to be called at the beginning of each event

  void VinciaWeights::resetWeights(int nAccepted) {

    for (int iVar=0; iVar<nWeightsSav; iVar++) {
      weights[iVar]     = 1.0;
      weightsOld[iVar]  = 0.0;
    }

    // Next doWeighting call will be the first one
    firstCall = true;

    // No reweighting so far
    didReweight = false;

    // Check if Pythia/Vincia vetoed event, restore counters
    if (nAccepted < nTotWeights) {
      nTotWeights--;
      nNonunityWeight        -= nNonunityWeightNow;
      nNegativeWeight        -= nNegativeWeightNow;
      nNonunityInitialWeight -= nNonunityInitialWeightNow;
      nNegativeInitialWeight -= nNegativeInitialWeightNow;

      for (int iVar=0; iVar<nWeightsSav; iVar++) {
        weightSum[iVar]  -= contribSum[iVar];
        weightSum2[iVar] -= contribSum2[iVar];
      }
    }

    nNonunityWeightNow        = 0;
    nNegativeWeightNow        = 0;
    nNonunityInitialWeightNow = 0;
    nNegativeInitialWeightNow = 0;

    for (int iVar=0; iVar<nWeightsSav; iVar++) {
      contribSum[iVar]  = 0.0;
      contribSum2[iVar] = 0.0;
    }

  }

  //--------------------------------------------------------------------------

  // Scaling functions

  // Scale the uncertainty band weights
  void VinciaWeights::scaleWeightVar(vector<double> Paccept, bool accept,
    bool isHard, string shower, int type) {
    if (!uncertaintyBands) return;
    if (nWeightsSav <= 1) return;
    if (!isHard && !varInMPIshowers) return;
    if (accept) scaleWeightVarAccept(Paccept,shower,type);
    else scaleWeightVarReject(Paccept,shower,type);
  }

  // Scale the uncertainty band weights if branching is accepted
  void VinciaWeights::scaleWeightVarAccept(vector<double> Paccept,
    string shower, int type) {
    for (int iVar=1; iVar<nWeightsSav; iVar++) {
      double PacceptVar = Paccept[iVar];
      if (PacceptVar > PACCEPTVARMAX) {
        PacceptVar = PACCEPTVARMAX;
        if (printPaccVarViol)
          infoPtr->errorMsg("Warning in VinciaWeights: resetting PacceptVar ("
            +shower+", "+num2str(type,2)+") for iVar = "+getWeightLabel(iVar));
      }
      scaleWeight( PacceptVar/Paccept[0], iVar );
    }
  }

  // Scale the uncertainty band weights if branching is rejected
  void VinciaWeights::scaleWeightVarReject(vector<double> Paccept,
    string shower, int type) {
    for (int iVar=1; iVar<nWeightsSav; iVar++) {
      double PacceptVar = Paccept[iVar];
      if (PacceptVar > PACCEPTVARMAX) {
        PacceptVar = PACCEPTVARMAX;
        if (printPaccVarViol)
          infoPtr->errorMsg("Warning in VinciaWeights: resetting PacceptVar ("
            +shower+", "+num2str(type,2)+") for iVar = "+getWeightLabel(iVar));
      }
      double reWeight = (1.0-PacceptVar)/(1.0-Paccept[0]);
      if (reWeight < MINVARWEIGHT) {
        reWeight = MINVARWEIGHT;
        if (printReWeightViol)
          infoPtr->errorMsg("Warning in VinciaWeights: resetting reWeight ("
            +shower+", "+num2str(type,2)+") for iVar = "+getWeightLabel(iVar));
      }
      scaleWeight( reWeight, iVar );
    }
  }

  //--------------------------------------------------------------------------

  // Main function to perform the reweighting due to MC overestimate
  // violations

  pair<double,double> VinciaWeights::doMCreweighting(vector<double>& Paccept) {

    // Save basic accept probability
    double PA = Paccept[0];
    for (int iVar=1; iVar<nWeightsSav; iVar++) Paccept[iVar] = PA;

    if (!allowReweighting) return make_pair(1.0,1.0);

    double acceptWeight = 1.0;
    double rejectWeight = 1.0;

    // f(t) = physical, g(t) = overestimate, h(t) = auxilary
    double physVSover = PA;         // = f(t)/g(t)
    double physVSaux  = physVSover; // = f(t)/h(t)
    double aFactor    = 1.0;        // = h(t)/g(t)

    // overestimate failed, choose h(t) = 1.1 f(t)
    if (physVSaux >= 1.0) {
      physVSaux = 1.0/1.1;          // = f(t)/h(t) = 1/1.1 < 1
      aFactor   = 1.1*physVSover;   // = h(t)/g(t) = 1.1 f(t)/g(t)
    }

    // negative, choose h(t) = - 1.1 f(t)
    else if (physVSover < 0.0) {
      physVSaux = 1.0/1.1;          // = f(t)/h(t) = 1/1.1 < 1
      aFactor   = 1.1*physVSover;   // = h(t)/g(t) = - 1.1 f(t)/g(t)
    }

    // accept weight = h(t)/g(t) [= 1 if overestimate ok]
    acceptWeight = aFactor;
    // reject weight = (1-f(t)/g(t))/(1-f(t)/h(t)) [= 1 if overestimate ok]
    rejectWeight = (1.0-physVSover)/(1.0-physVSaux);

    if (verbose >= 6) {
      printOut("VinciaWeights::doReweighting", "Reweighting event with:");
      cout << "  accept weight = " << acceptWeight << endl;
      cout << "  reject weight = " << rejectWeight << endl;
    }

    // Update accept probability (now < 1)
    Paccept[0] = physVSaux;
            PA = Paccept[0];
    for (int iVar=1; iVar<nWeightsSav; iVar++) Paccept[iVar] = PA;

    return make_pair(acceptWeight,rejectWeight);

  }

  //--------------------------------------------------------------------------

  // Helper function for keyword evaluation

  int VinciaWeights::doVarNow(string keyIn, int iAntPhys, string type) {
    string ASkey = ":murfac", NSkey = ":cns";
    // 1) Check variation for all branching types
    if (type + ASkey == keyIn) return 1;
    if (type + NSkey == keyIn) return 2;
    // 2) Check variation for specific branching type
    map<int,string> keyCvt = (type == "ff" ? iAntToKeyFSR : iAntToKeyISR);
    if (type + ":" + keyCvt[iAntPhys] + ASkey == keyIn) return 1;
    if (type + ":" + keyCvt[iAntPhys] + NSkey == keyIn) return 2;
    return -1;
  }

  //--------------------------------------------------------------------------

  // Main function to perform the uncertainty variations

  void VinciaWeights::doVariationsISR(vector<double>& Paccept, bool isHard,
    bool doMEC, BranchelementalISR* isrTrial, int iTrial, AlphaStrong* asPtr,
    double asUsr, double mu2usr, double mu2min, double mu2freeze, double asMax,
    bool isPolarized, double antUsr, double helSumUsr, double aHelUsr) {

    if (!uncertaintyBands) return;
    if (nWeightsSav <= 1) return;

    // Save basic accept probability
    double PA = Paccept[0];
    for (int iVar=1; iVar<nWeightsSav; iVar++) Paccept[iVar] = PA;

    if (!isHard && !varInMPIshowers) return;

    // Trial information
    double qBranch   = isrTrial->getTrialScale(iTrial);
    int    iAntPhys  = isrTrial->getPhysIndex(iTrial);
    bool   isII      = isrTrial->isII();

    // Invariants
    double sAnt = isrTrial->s();
    double s13  = 2.0*dot4(isrTrial->new1,isrTrial->new3);
    double s23  = 2.0*dot4(isrTrial->new2,isrTrial->new3);

    string type  = ( isII ? "ii" : "if" );

    // Loop over all variations
    for (int iVar=1; iVar<nWeightsSav; iVar++) {
      // Loop through keywords
      for (int iKey=0; iKey<(int)varKeys[iVar-1].size(); iKey++) {

        string keyNow = varKeys[iVar-1][iKey];
        double valNow = varVals[iVar-1][iKey];

        int varTypeNow = doVarNow(keyNow,iAntPhys,type);

        // 1) alphaS variation
        if (varTypeNow == 1) {
          double mu2var  = pow2(valNow)*mu2usr;
                 mu2var  = max(mu2min, mu2freeze + mu2var);
          double asVar   = min(asMax, asPtr->alphaS(mu2var));
          if (muSoftCorr && isrTrial->new2.id() == 21) {
            double zeta    = ( isII ? sAnt/s13 : sAnt/(sAnt+s23) );
            int    nf      = 5;
            if (qBranch < vinComPtr->mc) nf = 3;
            else if (qBranch < vinComPtr->mb) nf = 4;
            double b0      = (33.0 - 2.0 * nf) / (6.0 * M_PI);
            double mu2max  = max(sAnt, pow2(valNow)*mu2usr);
                   mu2max  = max(mu2min, mu2freeze + mu2max);
            double asCorr  = min(asMax, asPtr->alphaS(mu2max));
            double facCorr = 1.0 + (1.0 - zeta) * asCorr * b0 * log(valNow);
            asVar *= facCorr;
          }
          // Impose allowed range of variation
          if (abs(asVar-asUsr) > deltaASmax)
            asVar = (asVar>asUsr ? asUsr+deltaASmax : asUsr-deltaASmax);
          Paccept[iVar] *= asVar/asUsr;
        }

        // 2) Finite term variation
        else if ( !doMEC && (qBranch >= cNSpTmin) && (varTypeNow == 2) ) {
          double cNS     = valNow/sAnt;
          double antVar  = ant(antUsr,cNS);
          Paccept[iVar] *= antVar/antUsr;
          if (isPolarized) {
            double nHelConfig = 2.0;
            if ( iAntPhys == iGGemitIF || iAntPhys == iGQemitIF ||
              iAntPhys == iGQemitII ) nHelConfig = 3.0;
            else if ( iAntPhys == iGGemitII ) nHelConfig = 4.0;
            double aHelVar    = ant(aHelUsr,cNS/nHelConfig);
            double helSumVar  = max( ant(helSumUsr,cNS), aHelVar );
            Paccept[iVar]    *= (aHelVar/helSumVar)/(aHelUsr/helSumUsr);
          }
        }

      } // End loop through keywords
    } // End loop over variations

  }

  void VinciaWeights::doVariationsFSR(vector<double>& Paccept, bool isHard,
    bool doMEC, BranchelementalFSR* fsrTrial, int iTrial, AlphaStrong* asPtr,
    double asUsr, double mu2usr, double mu2min, double mu2freeze, double asMax,
    bool isPolarized, double antUsr, double helSumUsr, double aHelUsr) {

    if (!uncertaintyBands) return;
    if (nWeightsSav <= 1) return;

    // Save basic accept probability
    double PA = Paccept[0];
    for (int iVar=1; iVar<nWeightsSav; iVar++) Paccept[iVar] = PA;

    if (!isHard && !varInMPIshowers) return;

    // Trial information
    double qBranch   = fsrTrial->getTrialScale(iTrial);
    int    iAntPhys  = fsrTrial->getPhysIndex(iTrial);

    // Invariants
    double sAnt = fsrTrial->s();
    double s13  = 2.0*dot4(fsrTrial->new1,fsrTrial->new3);

    string type  = "ff";

    // Loop over all variations
    for (int iVar=1; iVar<nWeightsSav; iVar++) {
      // Loop through keywords
      for (int iKey=0; iKey<(int)varKeys[iVar-1].size(); iKey++) {

        string keyNow = varKeys[iVar-1][iKey];
        double valNow = varVals[iVar-1][iKey];

        int varTypeNow = doVarNow(keyNow,iAntPhys,type);

        // 1) alphaS variation
        if (varTypeNow == 1) {
          double mu2var  = pow2(valNow)*mu2usr;
                 mu2var  = max(mu2min, mu2freeze + mu2var);
          double asVar   = min(asMax, asPtr->alphaS(mu2var));
          if (muSoftCorr && fsrTrial->new2.id() == 21) {
            int    nf      = 5;
            if (qBranch < vinComPtr->mc) nf = 3;
            else if (qBranch < vinComPtr->mb) nf = 4;
            double b0      = (33.0 - 2.0 * nf) / (6.0 * M_PI);
            double mu2max  = max(sAnt, pow2(valNow)*mu2usr);
                   mu2max  = max(mu2min, mu2freeze + mu2max);
            double asCorr  = min(asMax, asPtr->alphaS(mu2max));
            double facCorr = 1.0 + (1.0 - s13/sAnt) * asCorr * b0 * log(valNow);
            asVar *= facCorr;
          }
          // Impose allowed range of variation
          if (abs(asVar-asUsr) > deltaASmax)
            asVar = (asVar>asUsr ? asUsr+deltaASmax : asUsr-deltaASmax);
          Paccept[iVar] *= asVar/asUsr;
        }

        // 2) Finite term variation
        else if ( !doMEC && (qBranch >= cNSpTmin) && (varTypeNow == 2) ) {
          double cNS     = valNow/sAnt;
          double antVar  = ant(antUsr,cNS);
          Paccept[iVar] *= antVar/antUsr;
          if (isPolarized) {
            double nHelConfig = 2.0;
            double aHelVar    = ant(aHelUsr,cNS/nHelConfig);
            double helSumVar  = max( ant(helSumUsr,cNS), aHelVar );
            Paccept[iVar]    *= (aHelVar/helSumVar)/(aHelUsr/helSumUsr);
          }
        }

      } // End loop through keywords
    } // End loop over variations

  }

  //--------------------------------------------------------------------------

  // Main function to perform the weighting

  void VinciaWeights::doWeighting() {

    double inW = 1.0;
    // Check for nonunity and negative initial weights
    if (firstCall) {
      inW = infoPtr->weight();
      bool report = false;
      bool isLast = false;
      if ( (abs(inW-1.0) > TINY) || (inW < 0.0) ) {
        nReportedWeight++;
        if (nReportedWeight <= nReportWeight) {
          report = true;
          if (nReportedWeight == nReportWeight) isLast = true;
        }
        string printMessage = "Nonunity initial weight occurred, w = ";
        // Updated counters
        if (inW < 0.0) {
          printMessage = "Negative initial weight occurred, w = ";
          nNegativeInitialWeight++;
          nNegativeInitialWeightNow++;
        } else {
          nNonunityInitialWeight++;
          nNonunityInitialWeightNow++;
        }
        // Always report in debugging mode
        if ((report && verbose >= 3) || verbose >= 4)
          printOut("VinciaWeights", printMessage + num2str(inW) + (isLast ?
            ": further output suppressed" : ""));
      }
    }

    // Add weight correction to cumulative sum (!= w0 for systems in series)
    // Take into account Pythia's weight (only != 1 for first time in event)
    double w0       = inW*weights[0];
    double wMC0     = weights[0];
    double wOld     = weightsOld[0];
    weightSum[0]   += (w0 - wOld);
    weightSum2[0]  += (pow2(w0) - pow2(wOld));
    // Contribution of current event
    contribSum[0]  += (w0 - wOld);
    contribSum2[0] += (pow2(w0) - pow2(wOld));

    if (firstCall) {
      firstCall = false;
      // Add one to number of events
      nTotWeights++;
    }

    // Check for nonunity and negative weights due to MC violations
    bool foundWeightToReport = false;
    if (abs(wMC0-1.0) > TINY) {
      // Only check if wOld = 0 or if weight changed
      if ( wOld < TINY || (abs(wOld) > TINY && (abs(wOld-w0) > TINY)) ) {
        didReweight         = true;
        foundWeightToReport = true;
        // Update counters
        nNonunityWeight++;
        nNonunityWeightNow++;
      }
    }
    if (wMC0 < 0.0 && wOld >= 0.0) {
      didReweight         = true;
      foundWeightToReport = true;
      // Update counters
      nNegativeWeight++;
      nNegativeWeightNow++;
    }
    bool report = false;
    bool isLast = false;
    if ( foundWeightToReport && (nReportedWeight <= nReportWeight) ) {
      report = true;
      if (nReportedWeight == nReportWeight) isLast = true;
    }
    // Always report in debugging mode
    if ((report && verbose >= 3) || verbose >= 4) {
      string printMessage = (wMC0 > 0.0 ? "Nonunity MC reweight occurred, w = "
        : "Negative MC reweight occurred, w = ");
      printOut("VinciaWeights", printMessage + num2str(w0) + (isLast ?
        ": further output suppressed" : ""));
    }

    // Keep track of minimum and maximum weights
    weightsMax[0] = max(weightsMax[0], w0);
    weightsMin[0] = min(weightsMin[0], w0);

    // Update the old weight
    weightsOld[0] = w0;

    // Save uncertainty band weights
    if (nWeightsSav>1) {
      for (int iVar = 1; iVar < nWeightsSav; iVar++) {
        double wVar        = inW*weights[iVar];
        double wVarOld     = weightsOld[iVar];
        // Add weight correction to cumulative sum
        weightSum[iVar]   += (wVar - wVarOld);
        weightSum2[iVar]  += (pow2(wVar) - pow2(wVarOld));
        // Contribution of current event
        contribSum[iVar]  += (wVar - wVarOld);
        contribSum2[iVar] += (pow2(wVar) - pow2(wVarOld));
        // Keep track of minimum and maximum weights
        weightsMax[iVar]   = max(weightsMax[iVar], wVar);
        weightsMin[iVar]   = min(weightsMin[iVar], wVar);
        // Update old weight
        weightsOld[iVar]   = wVar;
      }
    }

  }

}
